#!/usr/bin/env python
# -*- coding: iso-8859-1 -*-

import random
import itertools
import nltk
import sys
import numpy as np
import operator



porter = nltk.PorterStemmer()
EPSILON = 0.001
K = 10

def read_stopwords(path):
    """
    read stopword list
    """
    with open(path) as stream:
        lines = stream.readlines()
    stopwords = [l.rstrip().lower() for l in lines]
    return stopwords

STOPWORDS = set(read_stopwords("stopwords.txt"))


def answer_to_tokens(answer):
    """
    Given an answer, return a list of (preprocessed) tokens.
    """
    # split tokens: 'mind-blowing' -> 'mind', 'blowing'
    tokens = [t.split('-') for t in answer.split()]
    tokens = list(itertools.chain.from_iterable(tokens))

    # remove trailing or leading dots and apostrophs
    tokens = [t.strip('.') for t in tokens]
    tokens = [t.strip("'") for t in tokens]

    # stemming, lowercasing, stopword filtering
    for i,t in enumerate(tokens):
        try:
            tokens[i] = porter.stem(t)
        except:
            pass
    tokens = [t.lower() for t in tokens]
    tokens = [t for t in tokens if t not in STOPWORDS]

    return set(tokens)


def is_number(s):
    """
    returns True iff the given string is a number
    """
    try:
        _ = float(s)
        return True
    except:
        return False

    
def answer_to_number(answer):
    """
    An 'numeric' answer mainly contains a number, like 
    o December 26, 1991
    o 13,000 year ago
    o 277,923.1 km2.
    This method returns the last number contained
    in the answer, and True iff the answer appears
    to be 'numeric'.
    """
    answer = answer.replace('$', '') # $470
    answer = answer.replace(',', '') # 13,000 years
    tokens = answer.split()
    numbers = [t for t in tokens if is_number(t)]
    if len(tokens) <=3 and len(numbers) >= 1:
        return True, numbers[-1]
    else:
        return False, numbers[-1] if len(numbers)>0 else None



def compute_overlap_numeric(gt_answer, est_answer):
    """
    computes the overlap for numeric answers.
    The result is either 1 (iff ground truth answer and estimated
    answer contain the same number) or 0.
    """
    gt_isnumeric,gt_number = answer_to_number(gt_answer)
    est_isnumeric,est_number = answer_to_number(est_answer)
    if not gt_isnumeric:
        return 0.
    elif gt_number is not None and est_number is not None:
        return float(int(float(gt_number))==int(float(est_number)))
    else:
        return 0.



def compute_overlap_terms(gt_answer, est_answer):
    """
    computes a match score between two answers based on term overlap.
    The result is in the range [0,1], and is higher the more
    terms match.
    """
    tokens1 = answer_to_tokens(gt_answer)
    tokens2 = answer_to_tokens(est_answer)
    intersection = len(tokens1.intersection(tokens2))
    union = len(tokens1.union(tokens2))
    return float(intersection) / max(union, EPSILON)



def compute_overlap(gt_answer, est_answer):
    """
    Given a ground truth answer (gt_answer) and an
    estimated  answer (est_answer), compute the
    optimal overlap between the two (in percent).
    """
    return max(compute_overlap_terms(gt_answer, est_answer),
               compute_overlap_numeric(gt_answer, est_answer))



def eval(gt_answers, est_answer_lists):
        """
        This method measures how well a QA system's answers
        match the correct ("ground tuth") answers to a list of n 
        test questions.
        Given is 
        1. a list of n ground truth answers (gt_answers). 
        2. est_answer_lists, which contains n lists of 
           estimated answers (ranked, starting with the best).

        The method computes the mean reciprocal rank (MRR).
        It also computes a partial overlap between answers by two options:
        1. term overlap
        2. numeric overlap (e.g., "Dec 1991" matches "1991").

        @type gt_answers: list<string>
        @param gt_answers: The list of n answers to test questions 
                           (each a string)
        @type est_answer_lists: list<list<string>>
        @param est_answer_lists: A list of n elements. Each of those 
                                 contains a *list* of answers produced 
                                 by your QA system. This list is assumed 
                                 to be ranked, with the best answer 
                                 appearing as first, the second-best answer
                                 appearing as second, etc. 
                                 Only the top K of these answers will be
                                 used for evaluation.
        @returns: 1. the MRR score (our overall performance measure).
                  2. a list with each question's RR score.
                  3. each question's highest-scored answer. 
        @rtype: (float, np.array<float>, list<string>)
        """
        # consider only the top K answers
        est_answer_lists = [l[:K] for l in est_answer_lists] 
        all_scores = np.zeros(len(gt_answers))
        all_best_answers = [None] * len(gt_answers)

        for i,(gt_answer,est_answers) in enumerate(zip(gt_answers, est_answer_lists)):

            print('... comparing "%s" vs. "%s"' %(gt_answer, est_answers))

            if len(est_answers)==0:
                print('Warning: No answer for question no.', i, '(%s) !' %gt_answer)
                continue

            alphas = [compute_overlap(gt_answer,est_answer) for est_answer in est_answers]
            scores = [ 1./(j+1) * alpha for j,alpha in enumerate(alphas)]
            best   = np.argmax(scores) 

            all_scores[i] = scores[best]
            all_best_answers[i] = est_answers[best]

        return np.mean(all_scores), all_scores, all_best_answers



def small_test():
    gt_answers = ['John Wilkes Booth', '1947', '']    
    est_answer_lists = [ ['apple', 'cherry', 'J.W. Booth', 'John BOOTH'],
                         ['21 Jan 1948', '1948', 'Jan 1947', 'apple', 'cherry', 'pie', 'lord Buckethead'],
                         ['']
                       ]
    print ( eval(gt_answers, est_answer_lists) )



if __name__ == "__main__":

    small_test()
    
